# backup.md

%\red{
%Reading instruction:
%\begin{itemize}
%\item Intro has only pieces but should still give you the background.
%\item I have NOT polished Section 2-4.
%\item I have written and polished Section 5 but haven't got time to finish the explanation of the two counterexamples.
%\item You could skip Section 6; only bullets left there for the moment.
%\item I have NOT polished Section 7.
%\end{itemize}
%}

\begin{example}[Illustration of \si] \label{ex:banking}
  Consider the banking example of Figure~\ref{fig:banking}~\cite{Adya:PhDThesis1999, ClientCentric:PODC2017, AnalysingSI:JACM2018}.
  Alice and Bob share checking ($C$) and saving ($S$) accounts, each holding \$$50$.
  To maintain the invariant $C + S \ge 0$, before performing a withdrawal of $\mathit{amount}$,
  Alice and Bob check if $C + S \ge \mathit{amount}$ holds.
  Suppose Alice and Bob concurrently withdraw \$$60$
  from the checking and saving accounts in transactions $T_{A}$ and $T_{B}$, respectively.
  Under \si{}, both transactions are allowed to observe $C = S = 50$ such that $C + S \ge 60$ holds.
  Since $T_{A}$ and $T_{B}$ update different accounts,
  both can successfully commit, ending up with $C + S = -20$.
  This illustrates a non-serializable behavior called \emph{write skew}~\cite{Adya:PhDThesis1999, ClientCentric:PODC2017, AnalysingSI:JACM2018}.
\end{example}

  Consider the history of the banking example in Figure~\ref{fig:banking}.
  Since the dependency graph shown in the figure contains only cycles
  with two adjacent $\RW$ edges, i.e., $T_{A} \rel{\RW(S)} T_{B} \rel{\RW(C)} T_{A}$,
  this history satisfies \si.

% intro.tex
\begin{itemize}
  \item sound
  \item effective: find real bugs in production databases (confirmed by developers)
  \item efficiency
    \begin{itemize}
	  \item SI-specific optimizations
	    \begin{itemize}
		  \item search space pruning strategies
		\end{itemize}
	  \item \red{TODO: SMT solver-specific optimizations; in-monosat time comparison}
	    \begin{itemize}
		  \item encoding (setting decision variables)
		  \item deeply integrated with CDCL (e.g., learnt clauses)
		\end{itemize}
	\end{itemize}
  \item \red{TODO: informative}
    \begin{itemize}
	  \item minimal counterexamples
	  \item anomalies classification (Section 7.1 of \cite{ConsAD:VLDB2014})
	\end{itemize}
\end{itemize}

  % Let
  % \begin{align*}
	% \GraphSI = \set{\G \mid\; &(\T_{\G} \models \intaxiom) \;\land \\
	%   &(((\SO_{\G} \cup \WR_{\G} \cup \WW_{\G}) \comp \RW_{\G}?) \text{\it\; is acyclic})}.
  % \end{align*}
  % Then
  % \[
	%  \HistSI = \set{\H \mid \exists \WR, \WW, \RW.\; (\H, \WR, \WW, \RW) \in \GraphSI}.
  % \]

Then, we add the edges from $SC$ to the two graphs.
We encode the constraints in the following way:
Let $C = \tuple{E_{1}, E_{2}, t_{1}, t_{2}}$ be a constraint,
and $L_{1}, L_{2}$ be the sets of literals associated with the edges in $E_{1}$ and $E_{2}$.
Let
\begin{gather*}
  l_{1all} = \bigwedge_{l \in L_{1}}l \\
  l_{1none} = \bigwedge_{l \in L_{1}}\neg l \\
  l_{2all} = \bigwedge_{l \in L_{2}}l \\
  l_{2none} = \bigwedge_{\l \in L_{2}}\neg l \\
  l_{one} = (l_{1all} \wedge l_{2none}) \vee (l_{1none} \wedge l_{2all}).
\end{gather*}

$l_{one}$ means the edges from exactly one set in $C$ must exist in $G_{C}$. We
add $l_{one}$ to $L_{true}$ for all constraints.

The last step is to construct $GraphC$. We first add all edges in $GraphA$ to
it. Then, for every pair edges and their associated literals $<e_{a}, l_{a}>$ in
$GraphA$, $<e_{b}, l_{b}>$ in $GraphB$, if the destination of $e_{a}$ is the
source of $e_{b}$, we create a new edge $e_{c}$ in $GraphC$ as the concatnation
of $e_{a}$ and $e_{b}$, and its corresponding literal
$l_{c} = l_{a} \wedge l_{b}$. This is because $e_{c}$ exists in $G_{C}$ iff.
both $e_{a}$ and $e_{b}$ exists in the dependency graph.


% Systems which claims to support snapshot isolation:
% MongoDB, CockroachDB, PostgreSQL, Percolator, TiDB, Galera, DGraph, HBaseSI, Omid

t1 -so-> t2:
- still add constraints t1 -ww-> t2 or t2 -ww-> t1;
- will be pruned in the pruning phase

\begin{enumerate}
  \item \label{step:variables}
    For each vertex pair $(T_{i}, T_{j})$ ($i \neq j$),
    create a Boolean variable $E_{i, j}$. Add
	\begin{align}
	  \lnot E_{i, j} \lor \lnot E_{j, i}
	%   (E_{i, j} \implies \lnot E_{j, i}) \land (E_{j, i} \implies \lnot E_{i, j})
	  \label{encoding:at-most-one}
	\end{align}
	meaning that at most one of $E_{i, j}$ and $E_{j, i}$ can be \True.
	\begin{itemize}
	  \item Note that for vertices $i$ and $j$, we should consider both the edge $i \to j$ and the edge $j \to i$.
	  \item Note: it is ``at most one'', not ``exactly one''.
	    In other words, $E_{i, j}$ and $E_{j, i}$ can be \False{} at the same time.
	\end{itemize}
  \item \label{step:so-wr}
    Add all the $\SO$ edges and $\WR$ edges, by setting the corresponding
    Boolean variables to be \True.
    \begin{itemize}
	  \item Suppose that $i \rel{\SO \mid \WR} j$.
	    It is unnecessary to set $E_{j, i}$ to be \False.
	  \item Suppose that $i \rel{\SO \mid \WR} j$.
	    Do we really need to create a Boolean variable $E_{j, i}$ in step (1)?
	\end{itemize}
  \item \label{step:rw}
    Fix an object $x$.
    Consider any pair of transactions $T_{i}$ and $T_{j}$ ($i \neq j$) that
    both write on $x$ (that is, $T_{i} \rel{\WW_{x}} T_{j}$).
	\begin{enumerate}
	  \item Add
		\begin{align}
		E_{i, j} \lor E_{j, i}
		\label{encoding:at-least-one}
		\end{align}
		which, when combing with Eq. (\ref{encoding:at-most-one}),
		implies that exactly one of $E_{i, j}$ and $E_{j, i}$ can be \True.
		\begin{itemize}
		  \item \red{Question:} Enforce a total order among $\WriteTx_{x}$?
		    (SAT encoding??? Such subgraphs should be acyclic???)
		  \item \red{Question:} Is Eq. (\ref{encoding:at-least-one}) necessary?
		\end{itemize}
	  \item Let $\WR(x)(T_{i})$ (resp. $\WR(x)(T_{j})$) be the set of transactions
		that read $x$ from $T_{i}$ (resp. $T_{j}$).
		Add
		\begin{align}
		  \Big(E_{i, j} \implies \bigwedge_{T_{k} \in \WR(x)(T_{i})} E_{k, j}\Big) \land
		  \Big(E_{j, i} \implies \bigwedge_{T_{k} \in \WR(x)(T_{j})} E_{k, i}\Big)
		  \label{encoding:rw}
		\end{align}
	\end{enumerate}
\end{enumerate}
%%%%%%%%%%%%%%%%%%%%%%%%%


pruning-proof.tex
%%%%%%%%%%%%%%%
\begin{lemma}
  If there exists a cycle $c$, in which there exists an edge $e_{uv}$ $\in$ $C_{edge}$, then there must exist an edge $e_{xv} \in B_{edge}$, where $Last\_edge\_in\_A(vx) = true$.
\end{lemma}

\begin{proof}
  For such an edge $e_{uv}$, there must exist a path $p$ from $v$ to $u$. Since $e_{uv}$ $\in$ $C_{edge}$, there must exist another vertex $x$ and two edges $e_{ux}$ $\in$ $A_{edge}$ and $e_{xv}$ $\in$ $B_{edge}$. For edge $e_{xv}$ $\in$ $B_{edge}$ and there exists a path $p' = p + {e_{ux}}$ from $v$ to $x$, $Last\_edge\_in\_A(vx) = true$.
\end{proof}
%%%%%%%%%%%%%%%

\begin{figure}[t]
  \includegraphics[scale=0.4]{figs/completeness}
  \caption{Completeness of the pruning algorithm}
  \label{fig:pruning-completeness}
\end{figure}

%%%%%%%%%%%%%%%
\begin{lemma}
  If there exist an edge $e_{xv} \in B_{edge}$ and $Last\_edge\_in\_A(vx) = true$, then there exists a cycle $c$, in which there exists an edge $e_{uv}$ $\in$ $C_{edge}$.
\end{lemma}

\begin{proof}
  Since $Last\_edge\_in\_A(vx) = true$, there exists a path $p$ from $v$ to $x$ and the last edge $e_{ux}$ of $p$ is in $A_{edge}$. $e_{ux} \in A_{edge}$ and $e_{xv} \in B_{edge}$, so we can add an edge $e_{uv}$ $\in$ $C_{edge}$. Therefor cycle $c = p - e_{ux} + e_{uv}$ contains an edge $e_{uv}$ in $C_{edge}$.
\end{proof}
%%%%%%%%%%%%%%%

\begin{figure}[t]
  \includegraphics[scale=0.4]{figs/soundness}
  \caption{Soundness of the pruning algorithm}
  \label{fig:pruning-soundness}
\end{figure}

%%%%%%%%%%%%%%%
\begin{lemma}
  If there does not exist any edge $e_{xv} \in B_{edge}$ where $Last\_edge\_in\_A(vx) = true$, then there is not any cycle $c$, in which there exists an edge $e_{uv}$ $\in$ $C_{edge}$.
\end{lemma}

\begin{proof}
  This lemma is the converse negative proposition of lemma 1.
\end{proof}
%%%%%%%%%%%%%%%

% polygraph.tex (proof)
\begin{proof} \label{proof:polygraph-si}
  First, if $\H$ satisfies snapshot isolation,
  Then there must exists many possible version orders of it,
  and each version order of $\H$ is a directed graph.
  There must exist one version order
  which can construct a directed graph $G_1$,
  in which all cycles have at least two adjacent $RW$
  edges.\cite{AnalysingSI:JACM2018}.
  Since $G_1$ is constructed from a true version order,
  so it has the same known nodes and known edges of the polygraph of $\H$.
  $G_1$ also has one possible path of each constraint.
  We change the $RW$ edges into $C$ edges and construct a graph $G_2$,
  so $G_2$ is compatible with the polygraph of $\H$.
  For any cycle which has at least two adjacent $RW$ edges in $G_1$,
  the cycle will disappear in $G_2$.(~fig:\ref{fig:RW2C})
  So $G_2$ is acyclic.

  Second, if there exists an acyclic directed graph $G$
  that is compatible with the polygraph of $\H$,
  then $G$ has two kinds of edges and has no cycle.
  The first type of edges are the dependencies of $WW$, $SO$, $WR$.
  These edges are the same in $G$ and the polygraph of $\H$.
  The second type of edges are those `C edges'($WW/SO/WR + RW$),
  which is two adjacent edges in the poly graph.
  We change the `C edges' into two edges ($WW/SO/WR + RW$),
  and construct a new graph $G_1$, which still has no cycles.
  Here $G_1$ is a subgraph of the graph $G_2$
  constructed by a possible version order of $\H$,
  and $G_2/G_1$ only have some special $RW$ edges.
  Those $RW$ edges starts from the end of another $RW$ edge.
  Thus, those $RW$ edges can only create cycles
  has at least two adjacent $RW$ edges.
  So all cycles have at least two adjacent $RW$
  edges in $G_2$.
\end{proof}

% The inverse relation of $R$ is denoted by $R^{-1}$,
% i.e., $(a, b) \in R \iff (b, a) \in R^{-1}$.
% We use $R^{-1}(b)$ to denote the set $\{ a \in A \mid (a, b) \in R \}$.

To justify each transaction in a history,
we need to know how these transactions are related to each other.
This is captured declaratively
by the \emph{visibility} and \emph{arbitration} relations.

\begin{definition}[Abstract Executions] \label{def:ae}
  An \bfit{abstract execution} is a tuple $\ae = (\T, \SO, \vis, \ar)$,
  where $(\T, \SO)$ is a history,
  \bfit{visibility} $\vis \subseteq \T \times \T$ is a strict partial order,
  and \bfit{arbitration} $\ar \subseteq \T \times \T$
  is a strict total order such that $\vis \subseteq \ar$.
\end{definition}

For $\h = (\T, \SO)$, we often shorten $(\T, \SO, \vis, \ar)$ to $(\h, \vis, \ar)$.

\subsubsection{Consistency Axioms} \label{sss:axioms}

A consistency model is a set $\axiom$ of \emph{consistency axioms} constraining abstract executions.
The model allows those histories for which
there exists an abstract execution that satisfies the axioms:
$\hist_{\axiom} = \set{\h \mid \exists \vis, \ar.\; (\h, \vis, \ar) \models \axiom}$.

Table~\ref{table:axioms} summarizes all the consistency axioms necessary for \si.
For $T = (\E, \po)$, we let $T \vdash \writeevent(\keyxvar, \valvar)$
if $T$ writes to $\keyxvar$ and the last value written is $\valvar$,
and $T \vdash \readevent(\keyxvar, \valvar)$
if $T$ reads from $\keyxvar$ before writing to it
and $\valvar$ is the value returned by the first such read.
We also use $\WriteTx_{\keyxvar} = \set{T \mid T \vdash \writeevent(\keyxvar, \_)}$.
Two transactions $S$ and $T$ conflicts, denoted $S \conflict T$,
if they write on the same key.

The \emph{internal consistency axiom} \intaxiom{} ensures that,
within a transaction, a read from a key returns the same value
as the last write to or read from this key in the transaction.
The \emph{external consistency axiom} \extaxiom{} ensures that
an external read in a transaction $T$ from a key
returns the value written by the last transaction in $\ar$
among all the transactions that proceed $T$ in terms of $\vis$ and write this key.
The \extaxiom{} axiom implies the absence of
\emph{(1) dirty reads} ($\dirtyreads$):
a committed transaction cannot read a value from an aborted or ongoing transactions,
\emph{(2) intermediate reads} ($\intermediatereads$):
a transaction cannot read a value that was overwritten by the transaction that wrote it, and
\emph{(3) fractured reads} ($\fracturedreads$):
either all or none of the writes by a transaction can be visible to another transaction.
The \emph{prefix axiom} \prefixaxiom{} ensures that
if the snapshot taken by a transaction $T$ includes a transaction $S$,
than this snapshot also include all transactions that committed before $S$ in terms of $\ar$.
The \emph{no-conflict axiom} \noconflictaxiom{}
prevents concurrent transactions from writing on the same key.
%%%%%%%%%%%%%%%%%%%%
\subsubsection{Snapshot Isolation} \label{sss:si}

\si{} is formally defined as

\begin{definition}[$\si$~\cite{AnalysingSI:JACM2018}] \label{def:si}
  $\si = \intaxiom \land \extaxiom \land \sessionaxiom \land \prefixaxiom \land \noconflictaxiom$.
\end{definition}

Note that we are concerned with the Strong Session variant~\cite{LazyReplSI:VLDB2006, AnalysingSI:JACM2018} of \si{} with the $\sessionaxiom$ axiom,
which requires a transaction to observe all the transactions that precedes it in its session.
%%%%%%%%%%%%%%%%%%%%

% The \emph{external consistency axiom} \extaxiom{} ensures that
% an external read in a transaction $T$ from a key
% returns the value written by the last transaction in $\ar$
% among all the transactions that proceed $T$ in terms of $\vis$ and write this key.
% The \extaxiom{} axiom implies the absence of

% \emph{(3) fractured reads} ($\fracturedreads$):
% either all or none of the writes by a transaction can be visible to another transaction.
% Note that the $\fracturedreads$ property which is also implied by the $\extaxiom$ axiom
% involves the unknown $\WW$ dependencies and is not trivially checkable.

For some subset $A' \subseteq A$, the restriction of $R$ to $A'$ is
$R|_{A'} \triangleq R \cap (A' \times A')$.

To cope with the \np-hardness of the \si{} checking problem,
we reduce it to a constraint satisfaction problem on graphs
and utilize MonoSAT, an off-the-shelf SMT solver
well-suited to checking graph properties such as acyclicity, to solve it.
The theoretical foundation of this reduction is the \emph{dependency} graph-based
characterization of \si{} in Theorem~\ref{thm:depgraph-si}.

% To capture the constraints on \emph{all} possible dependencies between transactions,
% in this section, we present a new \emph{polygraph} based characterization of \si{}.
% \red{Section~\ref{section:si-sat}} describes the way how these constraints can be encoded
% into SAT formulas for MonoSAT to solve efficiently.

%\noindent These data consistency properties are guaranteed by the weaker isolation levels than SI:
% $\intaxiom$ by \emph{cut isolation} \cite{ra},
% $\dirtyreads$ and $\intermediatereads$ by \emph{read committed}~\cite{CritiqueANSI:SIGMOD1995},
% and $\thinairreads$ by \emph{transactional causal consistency}~\cite{DBLP:conf/popl/BouajjaniEGH17,eiger}.

%%%%%%%%%%%%%%%
% \begin{figure}[t]
%   \includegraphics[width = 0.48\textwidth]{figs/RW2C}
%   \caption{Breaking cycles with adjacent RW edges.}
%   \label{fig:RW2C}
% \end{figure}
%%%%%%%%%%%%%%%


      \red{
      \ForAll{$T, S \in \T$ such that $T \neq S$}
      \label{line:coalesce}
      \State $\eithervar \gets \set{(T, S, \WW)}$
      \State $\orvar \gets \set{(S, T, \WW)}$
      \ForAll{$x \in \text{WriteKey}_{T} \cap \text{WriteKey}_{S}$}
        \State $\eithervar \gets @ \cup \bigcup\limits_{T' \in \WR(\keyxvar)(T)} \!\!\set{(T', S, \RW)}$
        \State $\orvar \gets @ \cup \bigcup\limits_{S' \in\WR(\keyxvar)(S)} \!\!\set{(S', T, \RW)}$
        \EndFor
        \State $\cons_{\g} \gets @ \cup \set{\tuple{\eithervar, \orvar}}$
      \EndFor
      }

% This kind of generalized constraints brings two benefits to later optimizations:

% To speed up MonoSAT, \name{} explicitly specify the decision variables when calling MonoSAT,
% saving the CDCL~\cite{} search procedure of MonoSAT from deciding on variables
% that are implied by the decision variables.
% Specifically, we specify the set of unknown Boolean variables associated with $\WW$ edges
% in the constraints $\cons_{\g}$ as the only decision variables in the MonoSAT solver
% (lines~\code{\ref{alg:checksi}}{\ref{line:solveconstraints-ww}}
% and~\code{\ref{alg:checksi}}{\ref{line:solveconstraints-dv}}).
% The set of unknown Boolean variables associated with $\RW$ edges in $\cons_{g}$
% will be forced by \emph{unit propagation} in the CDCL search procedure of MonoSAT.
% For example, consider a clause of the form
% $\BV_{i, j} \implies \bigwedge_{(\vvar_{k}, \vvar_{j}, \RW) \in \eithervar \setminus \set{\evar}} \BV_{k, j}$
% at line~\code{\ref{alg:checksi}}{\ref{line:encodeconstraints-clauses-add-implication-for-either}}.
% If MonoSAT decides to set the decision variable $\BV_{i, j}$ on the left hand to be \True,
% then all the non-decision variables $\BV_{k, j}$ on the right hand must be also \True.
% This can be done via unit propagation, saving MonoSAT from trying both cases for each $\BV_{k, j}$ variable.

Criteria:
\begin{itemize}
  \item efficiency
    \begin{itemize}
	  \item SI-specific optimizations
	    \begin{itemize}
		  \item search space pruning strategies
		\end{itemize}
	  \item \red{TODO: SAT solver-specific optimizations; in-monosat time comparison}
	    \begin{itemize}
		  \item encoding (setting decision variables)
		  \item deeply integrated with CDCL (e.g., learnt clauses)
		\end{itemize}
	\end{itemize}
  \item \red{TODO: informative}
    \begin{itemize}
	  \item minimal counterexamples
	  \item anomalies classification (Section 7.1 of \cite{ConsAD:VLDB2014})
	  \item \nobi{revisit our meeting on July 1st;  think also about the Galera-Maria case study: how we can derive the lost update anomaly from the counterexample returned by our alg. In other words, the returned counterexample says nothing about lost update;  how can we automate our manual identification of such anomaly? I think this is also relevant to ``minimal counterexamples''.}
	\end{itemize}
\end{itemize}

\begin{figure*}[t]
  \centering
  \includegraphics[scale=0.8]{figs/bug_dgraph.png}
  \caption{Illustrating Algorithm \ref{} using the new SI violation found in DGgraph.}
  \label{fig:ce-dgraph}
\end{figure*}

\begin{itemize}
  \item (MonoSAT Encoding) According to the Serializability theorem~\cite{},
    a history $\h$ satisfies \ser{} if and only if
	there exists a \red{commit order} such that
	the dependency graph arising from $H$ and that commit order is acyclic.
	This acyclicity can be directly encoded into MonoSAT.
	In contrast, the \si{} Characterization Theorem~\cite{AnalysingSI:JACM2018} state that
	a history $h$ satisfies \si{} if and only if
	there exists a commit order such that
	the dependency graph arising from $H$ and that commit order
	contains only cycles with at least two adjacent anti-dependency edges.
	It is non-trivial to encode such (a)cyclicity in MonoSAT.
  \item (Pre-MonoSAT Optimization)
    pruning types of edges matter; not the case in \cobra
\end{itemize}

SMT solver well-suited to checking graph properties such as acyclicity,

% \red{More importantly, in Section~\ref{ss:counterexample} we will utilize the fact that transactions are grouped into sessions
% to reduce the size of the witness of the \si{} violation returned by \name.}


      \ForAll{$\tuple{\eithervar, \orvar} \in \cons_{\g}$}  \Comment{encode the constraints}
        \label{line:encodeconstraints-forall-cons}
        \State choose the unique $\evar \gets (\vvar_{i}, \vvar_{j}, \WW) \in \eithervar$
          \label{line:encodeconstraints-choose-ww-in-either}
        \State $\Clause \gets @ \cup \set{\BV_{i, j} \implies
          \bigwedge\limits_{(\vvar_{k}, \vvar_{j}, \RW) \in \eithervar \setminus \set{\evar}} \BV_{k, j}}$
          \label{line:encodeconstraints-clauses-add-implication-for-either}
        \State choose the unique $\evar' \gets (\vvar_{j}, \vvar_{i}, \WW) \in \orvar$
          \label{line:encodeconstraints-choose-ww-in-or}
        \State $\Clause \gets @ \cup \set{\BV_{j, i} \implies
          \bigwedge\limits_{(\vvar_{k}, \vvar_{i}, \RW) \in \orvar \setminus \set{\evar'}} \BV_{k, i}}$
          \label{line:encodeconstraints-clauses-add-implication-for-or}
        \State $\Clause \gets @ \cup \set{\BV_{i, j} \lor \BV_{j, i}}$
          \label{line:encodeconstraints-clause-bvij-or-bvji}
      \EndFor


      % \ForAll{$\vvar_{i}, \vvar_{j} \in \vertex_{\g}$ such that $i < j$}
      %   \Comment{\red{unnecessary? for acyclicity?}}
      %   \State $\Clause \gets @ \cup \set{\lnot \BV_{i, j} \lor \lnot \BV_{j, i}}$
      %     \Comment{\red{not in \cobra?}}
      %     \label{line:encodeconstraints-cl}
      % \EndFor


% \red{Consider a
% Let $\evar \gets (\vvar_{i}, \vvar_{j}, \WW)$ be the unique $\WW$ edge in $\eithervar$
% (line~\code{\ref{alg:checksi}}{\ref{line:encodeconstraints-choose-ww-in-either}}).
% We add the clause $\BV_{i, j} \implies
% \bigwedge_{(\vvar_{k}, \vvar_{j}, \RW) \in \eithervar \setminus \set{\evar}} \BV_{k, j}$ to $\Clause$
% (line~\code{\ref{alg:checksi}}{\ref{line:encodeconstraints-clauses-add-implication-for-either}}),
% enforcing that if $\evar$ is included in the resulting compatible graph,
% then so do all the other edges in $\eithervar$.
% The possibility of $\orvar$ is encoded similarly
% (lines~\code{\ref{alg:checksi}}{\ref{line:encodeconstraints-choose-ww-in-or}}
% and \code{\ref{alg:checksi}}{\ref{line:encodeconstraints-clauses-add-implication-for-or}}).
% The clause $\BV_{i, j} \lor \BV_{j, i}$ at line~\code{\ref{alg:checksi}}{\ref{line:encodeconstraints-clause-bvij-or-bvji}}
% ensures that \red{at least one} of the two possibilities of $\eithervar$ and $\orvar$ happens.

  \item \red{\emph{thin-air reads}} ($\thinairreads$):
    a transaction cannot read a thin-air value, i.e., one that does not appear in the history.

% \begin{theorem}[Completeness of \name] \label{thm:completeness}
%   Assume that the status of transactions in every history is definitely committed or aborted.
%   Then \name{} is complete with respect to a given history, i.e.,
%   if the history violates \si, \name{} returns \false.
% \end{theorem}

% \begin{proof} \label{proof:completeness}
%   \red{TODO:}
% \end{proof}

	% Thanks to MonoSAT, \name{} identifies the \si{} violations as a pair of undesired cycles
	% in dependency graphs of the input history.
	% \name{} further reduces the size of the cycles \red{so that $\dots$},
	% greatly easing the understanding of the violations.


The key to solving the \si{} checking problem with respect to a history
is to resolve the \emph{uncertainties} on the version orders between all transactions that write the same key in the history.
Biswas and Enea showed that when the version orders for each key are unknown in the given history,
the \si{} checking problem is \npc~\cite{Complexity:OOPSLA2019}.
In contrast, the \si{} checking problem can be easily solved in time polynomial in the size of the history
if the version orders for each key are known in a history.
Biswas and Enea further proposed a polynomial-time algorithm for checking \si{}
when the number of sessions is considered to be a constant.
However, it is neither practically efficient
(see our performance evaluation in Section~\ref{section:experiments}) nor informative.

Elle is also able to check \si{} (and other isolation levels) on histories in polynomial time~\cite{Elle:VLDB2020}.
Although Elle is quite efficient,
it relies on specific data models, e.g., lists, in workloads
to infer the version orders for each key from histories.
Even more problematic, Elle relies on the start and commit timestamps of transactions,
which may not be exposed to clients, to infer the \emph{start-dependency relation} used in Adya's formalization of \si.

A third approach, called CobraSI, reduces the history for \si{} checking to that for \ser{} checking
using the polynomial-time reduction procedure proposed in~\cite{Complexity:OOPSLA2019},
and then solves it utilizing Cobra~\cite{Cobra:OSDI2020}, a state-of-the-art serializability checker.
However, this approach is neither practically efficient due to the reduction procedure
(see our performance evaluation in Section~\ref{section:experiments}) nor informative due to Cobra.

https://fauna.com/blog/introduction-to-transaction-isolation-levels

Implementation of the IC3 / Property Directed Reachability algorithm using the the Z3 SMT solver.

\begin{itemize}
  \item (Characterization of \si) The uncertainties on the version orders for each key in a history
	can be captured by the dependency graphs of this history~\cite{Adya:PhDThesis1999}.
	Roughly speaking, a history satisfies \si{} if and only if
	there \emph{exists} a dependency graph of this history such that
	it contains only cycles of a specific pattern~\cite{AnalysingSI:JACM2018}.
	However, a dependency graph of a history represents only a single possibility of the version orders.
	How should we characterize the uncertainties on version orders
	in a single structure so that we can encode them all together?
  \item (Efficient) In the worst case, the number of different possible version orders in a history is exponential in the size of the history.
	It is unrealistic to enumerate and check them individually in a brute-force manner.
	What is an efficient way to resolve the uncertainties on potential version orders
	and even better prune the impossible ones in a history?
  \item (Informative) How to also identify the representative violations of \si{}
	if a history does not satisfy \si?
	Furthermore, how to help developers understand the violations, especially when
	the violations involve many transactions?
\end{itemize}

% A directed graph $G = (V, E)$ is a pair consisting of
% a set $V$ of vertices and a set $E \subseteq V \times V$ of edges.
A directed labeled graph $G = (V, E)$ is a pair consisting of
a set $V$ of vertices and a set $E \subseteq V \times V \times \lbl$ of edges, where $\lbl$ is a set of labels.
For a directed labeled graph $G$ and a label $L \in \lbl$,
we use $V_{G}$, $E_{G}$, and $L_{G}$ to denote the set of vertices, the set of edges,
and the set of edges with label $L$ in $G$, respectively.
For a labeled graph $G = (V, E)$ and a set $F$ of edges,
we use \red{$G|_{F}$} to denote the graph that has the set $V$ of vertices and the set $F$ of edges.
% A label $L \in \lbl$ induces a relation defined as $L \triangleq \set{(a, b) \mid (a, b, L) \in E}$.
% We thus define compositions of edges in terms of relation compositions:
% For any two labels $L_{1}$ and $L_{2}$ and
% any subsets $R \subseteq L_{1}$ and $S \subseteq L_{2}$, $R \comp S$
For any two subsets $R, S \subseteq E$, we define their composition as
$R \comp S = \{ (a, c, L_{1} \comp L_{2}) \mid
  \exists b \in V: (a, b, L_{1}) \in R \land (b, c, L_{2}) \in S \}$,
where $L_{1} \comp L_{2}$ is a newly introduced label when composing edges.
% That is, we ignore labels when composing edges.

\begin{theorem}[] \label{thm:preserving-si-acyclicity}
  Let $G$ and $G_{p}$ be the generalized polygraphs
  just before and after \emph{$\pruneconstraints$}, respectively.
  $G$ is \si-acyclic if and only if \emph{$\pruneconstraints$} returns \emph{\True{}} and $G_{p}$ is \si-acyclic.
\end{theorem}

\begin{theorem}[No New Undesired Cycles] \label{thm:pruning-no-new-cycles}
  Let $G$ and $G_{p}$ be the generalized polygraphs
  just before and after \emph{$\pruneconstraints$}, respectively.
  Suppose that $G_{p}$ is not \si-acyclic.
  Let $\cycle$ be a cycle in a compatible graph with the induced \si{} graph of $G_{p}$.
  Then there is a compatible graph with the induced \si{} graph of $G$ that contains $\cycle$.
\end{theorem}

% Formally, $\forall \keyxvar \in \Key.\; \forall \opvar_{1}, \opvar_{2} \in \Op.\; \forall \valvar_{1}, \valvar_{2} \in \Val.\;
%   (\opvar_{1} = \writeevent(\keyxvar, \valvar_{1}) \land \opvar_{2} = \writeevent(\keyxvar, \valvar_{2})
%     \land \opvar_{1} \neq \opvar_{2}) \implies \valvar_{1} \neq \valvar_{2}$.

% For a set $A \neq \emptyset$ and a strict total order $R \subseteq A \times A$,
% we let $\max_{R}(A)$ be the element $a \in A$ such that $\forall b \in A.\; b = a \lor (b, a) \in R$.

% In the following, we describe these steps in details.
% For conciseness, we write expressions like $S \gets @ \oplus T$
% to denote $S \gets S \oplus T$, where $\oplus$ is a binary operator.

%counter_example_tikz.tex

%\begin{figure*}[t]
%  \centering
%    \begin{subfigure}[c]{0.3\textwidth}
%        \centering
%        \includegraphics[width = 1\textwidth]{figs/Galera_counter_example_original}
%        \caption{The original output of the violation in Galeta.}
%    \label{fig:original-galera}
%    \end{subfigure}
%    \hfill
%    \begin{subfigure}[c]{0.3\textwidth}
%        \centering
%        \includegraphics[width = 1.00\textwidth]{figs/Galera_counter_example_recovery}
%        \caption{Recover the information.}
%        \label{fig:recover-galera}
%    \end{subfigure}
%    \hfill
%    \begin{subfigure}[c]{0.3\textwidth}
%        \centering
%        \includegraphics[width = 1.00\textwidth]{figs/Galera_counter_example_recovery}
%        \caption{Delete uncertainty.}
%        \label{fig:delete-galera}
%    \end{subfigure}
%    \label{fig:counter_example_galera}
%    \caption{Counter example of Galera}
%\end{figure*}

%\begin{figure*}[t]
%  \centering
%    \begin{subfigure}[c]{0.2\textwidth}
%        \centering
%        \includegraphics[width = \textwidth]{figs/DGraph_counter_example_original}
%        \caption{The original output of the violation in DGraph.}
%    \label{fig:original-DGraph}
%    \end{subfigure}
%    \hfill
%    \begin{subfigure}[c]{0.3\textwidth}
%        \centering
%        \includegraphics[width = 1.3\textwidth]{figs/DGraph_counter_example_recovery}
%        \caption{Recover the information.}
%        \label{fig:recover-DGraph}
%    \end{subfigure}
%    \hfill
%    \begin{subfigure}[c]{0.3\textwidth}
%        \centering
%        \includegraphics[width = 1.3\textwidth]{figs/DGraph_counter_example_recovery}
%        \caption{Delete uncertainty.}
%        \label{fig:delete-DGraph}
%    \end{subfigure}
%    \label{fig:counter_example_DGraph}
%    \caption{Counter example of DGraph}
%\end{figure*}

%\begin{figure*}[t]
%  \centering
%    \begin{subfigure}[c]{0.3\textwidth}
%        \centering
%        \includegraphics[width = 1.00\textwidth]{figs/Yuga_counter_example_original}
%        \caption{Original output}
%    \label{fig:original-Yuga}
%    \end{subfigure}
%    \hfill
%    \begin{subfigure}[c]{0.3\textwidth}
%        \centering
%        \includegraphics[width = 1.00\textwidth]{figs/Yuga_counter_example_recovery}
%        \caption{Recovered  scenario}
%        \label{fig:recover-Yuga}
%    \end{subfigure}
%    \hfill
%    \begin{subfigure}[c]{0.3\textwidth}
%        \centering
%        \includegraphics[width = 1.00\textwidth]{figs/Yuga_counter_example_recovery}
%        \caption{Finalized  scenario}
%        \label{fig:delete-DGraph}
%    \end{subfigure}
%    \label{fig:counter_example_Yuga}
%    \caption{Causality violation: the SI anomaly found in YugabyteDB}
%\end{figure*}